#!/usr/bin/perl -w
#
# QualiTree.pl v1.2.1
# Last modified April 18, 2007 11:09
# (c) Olaf R.P. Bininda-Emonds
#
# Requirements:
#	A nexus-formatted treefile in which the *first* tree is the supertree and all
#	remaining trees are the source trees. The trees can be in different tree blocks
#	and translation tables are also acceptable. However, all trees must be in the
#	same file.
#
# Generates:
#   1) Tab-delimited files (QStree.txt and QSclade.txt) describing qualitative
#      support for the clades in a supertree relative to the clades in the source trees
#	2) Nexus-formatted treefile with QS values contained as node labels (can be viewed, for
#	   example in TreeView)
#
# Usage: QualiTree.pl -t<filename> [-a] [-b] [-n<replicate number>] [-r] [-v] [-h]
#   options: -a = append to output files (QS_tree.txt, QS_clade.txt, QS_supertree.tre, and, optionally, QS_bootstrap.txt)
#            -b<number> = number of replicates for bootstrap resampling of source trees
#                         (default = 0 (off); suggested level = 10000)
#            -n<replicate number> = replicate number (for simulated data)
#                                   (default = 1)
#            -r = do not perform reduced analysis where supertree is pruned to match source tree taxon set
#			 -s = do not give statistics for each clade in the supertree (to file QS_clade.txt)
#            -t<filename> = treefile containing trees in NEXUS format
#            -v = verbose output
#            -h = print this message and quit

use strict;
use Math::Trig;
use File::Basename;

# Set defaults
	# I/O variables
		my $treefile;
		my $block = 1;
		my $treeBlock = 0;
		my $bootRep = 0;
		my $reducedAnalysis = 1;
		my $statAnalysis = 1;
	
	# Tree variables
		my $SupportFlag;

	# Output variables
		my $write = ">";
		my ($outFile1, $outFile2, $outFile3, $outFile4);
			my %abbrevStatus = ('hardSupport' => 'HS', 'softSupport' => 'SS', 'hardConflict' => 'HC', 'softConflict' => 'SC', 'equivocal' => 'EQ');
			my %nodeLabel;
			my %QSclade;
	
	# Miscellaneous variables
		my $iRep = 1;
		my ($startTime, $totalTime);
		my $verbose = 0;
		my $version = "1.2.1";
		
# Check switches
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] eq "-a")
			{
			$write = ">>";
			}
		elsif ($ARGV[$i] =~ /-b([\d]*)/)
			{
			$bootRep = $1;
			}
		elsif ($ARGV[$i] =~ /-n([\d]*)/)
			{
			$iRep = $1;
			}
		elsif ($ARGV[$i] eq "-r")
			{
			$reducedAnalysis = 0;
			}
		elsif ($ARGV[$i] eq "-s")
			{
			$statAnalysis = 0;
			}
		elsif ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
			{
			$treefile = $1;
			die "ERROR: Treefile $treefile does not exist.\n" unless (-e $treefile);
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "Usage: QualiTree.pl -t<filename> [-a] [-b] [-n<replicate number>] [-r] [-s] [-v] [-h]\n";
			print "Version: $version\n";
			print "Options: -a = append to output files (QS_tree.txt, QS_clade.txt, QS_supertree.tre)\n";
			print "              and, optionally, QS_bootstrap.txt\n";
			print "         -b<number> = number of replicates for bootstrap resampling of source trees\n";
			print "                      (default = 0 (off); suggested level = 10000)\n";
			print "         -n<replicate number> = replicate number (for simulated data)\n";
			print "                                (default = 1)\n";
			print "         -s = do not give statistics for each clade in the supertree (to file QS_clade.txt)\n";
			print "         -r = do not perform reduced analysis where supertree is pruned to match source tree taxon set\n";
			print "         -t<filename> = file containing trees in NEXUS format\n";
			print "         -v = verbose output\n";
			print "         -h = print this message and quit\n";
			print "\nQS citation:\n\tBininda-Emonds, O.R.P. 2003. Novel versus unsupported clades: assessing the qualitative support for clades in MRP supertrees. Systematic Biology 52(6):839848.\n";
			print "\nrQS citation:\n\tPrice, S.A., O.R.P. Bininda-Emonds, and J.L. Gittleman. 2005. A complete phylogeny of the whales, dolphins and even-toed hoofed mammals (Cetartiodactyla). Biological Reviews 80(3):445473.\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: QualiTree.pl -t<filename> [-a] [-b] [-n<replicate number>] [-r] [-s] [-v] [-h]\n";
			print "Version: $version\n";
			exit(1); 
			}
		}
	
# Check for necessary conditions and set output files
	if (not defined $treefile)
		{
		die "ERROR: Must specify treefile containing supertree and source tree descriptions.\n";
		}
	else
		{
		my ($dataSource, $dir, $ext) = fileparse($treefile, qr/\.\D.*/);
			$outFile1 = $dataSource . "_rQS_tree.txt";
			$outFile2 = $dataSource . "_rQS_node.txt";
			$outFile3 = $dataSource . "_rQS_supertree.tre";
			$outFile4 = $dataSource . "_rQS_bootstrap.txt";
			
		unless ($reducedAnalysis)
			{
			$outFile1 =~ s/_rQS/_QS/;
			$outFile2 =~ s/_rQS/_QS/;
			$outFile3 =~ s/_rQS/_QS/;
			$outFile4 =~ s/_rQS/_QS/;
			}
		}

print "Reading in tree descriptions ...\n";
	$startTime = time;
	setLineBreak($treefile);
	open (fTREES, "<$treefile") or die "Cannot open treefile $treefile\n";		# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
		my (@AllTrees, %treeNames, $treeDescrip, $treeFlag, $treeCount);
		my (@taxonList, %taxonPresent);
		my $NumChar = 0;
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<fTREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$treeDescrip .= $_;
				if ($_ =~ m/:/)
					{
					$SupportFlag = 1;
					}
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		#Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		#Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		#Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						push @AllTrees, $tree;
					print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close fTREES;

	$totalTime = time - $startTime;
	print "\tTime taken: $totalTime seconds\n";
	die "\nERROR: No trees read in from file $treefile\n" if (not @AllTrees);

my @AllTaxa = sort {$a cmp $b} @taxonList;
	unshift @AllTaxa, "MRPoutgroup";
	my $NumTaxa = scalar(@AllTaxa);

print "\nParsing tree descriptions ...\n";
	$startTime = time;
	$treeCount = 0;
	my $NumElement = 0;
	my (@element, @CladeSupport, @STsupport, @superTree, @STcladeID, %sourceTree);
	my (%ParentClade, %CladeDescrip, %CladeSupport, %open, %start, %stop, %present);
	my (%treeStop);
	my $charCount = 1;

	foreach my $tree (@AllTrees)	# Get structure of each tree
		{
		$treeCount++;
		my %present;
		print "\nTree: $treeCount\n" if ($verbose);
		my $totalClade = ($tree =~ tr/(//);
		$treeStop{$treeCount} = $charCount + $totalClade - 2;
		my $charCount = $treeStop{$treeCount} + 1;
		my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
		my $OldCurrent;
		
		$tree =~ s/:\d+\.?\d*e?\-?\d*$//ig;	# Remove trailing branch length information
		
		while ($position < length($tree))
			{
			my $offset = 1;
			if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$CladeID++;
				if ($CladeID > 1)
					{
					$NumElement++;
					$element[$CladeID] = $NumElement;
					}
				$ParentClade{$CladeID} = $CurrentClade;
				$CurrentClade = $CladeID;
				$open{$CurrentClade} = 1;
				$start{$CladeID} = $position;
				}
			elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$open{$CurrentClade} = 0;
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
				$stop{$OldCurrent} = $position;
				$CladeDescrip{$OldCurrent} = substr($tree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
				$CladeDescrip{$OldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
				}
			elsif (substr($tree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or $position+$offset >= length($tree))
					{
					$offset++;
					}
				if (substr($tree, $position-1, 1) eq ")")	# Support value pertains to a clade
					{
					$CladeSupport[$element[$OldCurrent]] = substr($tree, $position+1, $offset-1);
					if ($treeCount == 1 and $SupportFlag)
						{
						$STsupport[$OldCurrent-2] = substr($tree, $position+1, $offset-1);
						}
					}
				$position += $offset-1;
				}
			elsif (substr($tree, $position, 1) eq ",")
				{
				}
			else	#Determine membership of clade
				{
				until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":")
					{
					$offset++;
					}
				for (my $clade = 1; $clade <= $CladeID; $clade++)	# Assign taxon to any open clades
					{
					if ($open{$clade})
						{
						$present{substr($tree, $position, $offset)}{$clade} = 1;
						}
					}
				$position += $offset-1;
				}
			$position++;
			}
			
		if ($verbose)
			{
			for (1..$totalClade)
				{
				print "\t$_\t$CladeDescrip{$_}\n";
				if ($_ > 1)
					{
					print "\t\tSupport: $CladeSupport[$element[$_]]\n" if ($CladeSupport[$element[$_]]);
					}
				}
			}
		
		for (my $CladeCount = 1; $CladeCount <= $totalClade; $CladeCount++)	#MRP coding subroutine in transposed format
			{
			my $MatrixRep;
			for (my $TaxonCount = 0; $TaxonCount < $NumTaxa; $TaxonCount++)
				{
				if ($TaxonCount == 0)	#Hypothetical outgroup; coded as 0 for every tree
					{
					$MatrixRep .= "0";
					}
				elsif ($present{$AllTaxa[$TaxonCount]}{1})	#Taxon is present on this source tree; {$clade = 1} is basal clade of tree
					{
					if ($present{$AllTaxa[$TaxonCount]}{$CladeCount})		#Taxon is present in focal clade
						{
						$MatrixRep .= "1";
						}
					else
						{
						$MatrixRep .= "0";
						}
					}
				else	#Taxon is missing from this source tree
					{
					$MatrixRep .= "?";
					}
				}
			if ($treeCount == 1)
				{
				push @superTree, $MatrixRep;
				my $StripClade = $CladeDescrip{$CladeCount};
					$StripClade =~ s/\(//g;
					$StripClade =~ s/\)//g;
					$StripClade = "(".$StripClade.")";
				push @STcladeID, $StripClade;
				}
			else
				{
				push @{$sourceTree{$treeCount-2}}, $MatrixRep;
				}
			}
		}

	$totalTime = time - $startTime;
	print "\tTime taken: $totalTime seconds\n";

my $NumSource = $treeCount - 1;

open (OUT, $write, $outFile1) or die "Cannot open output file $outFile1\n";
open (NODE, $write, $outFile2) or die "Cannot open output file $outFile2\n";

if ($iRep == 1)
	{
	print OUT "\t\tQSindex(tree)\n";
	print OUT "Replicate\tNumClades\t(mean)\t(variance)\t";
	print OUT "NumHardMatch\t";
		print OUT "aveSupportHardSupport\t" if (@STsupport);
		print OUT "aveCladeSizeHardSupport\t";
	print OUT "NumSoftMatch\t";
		print OUT "aveSupportSoftSupport\t" if (@STsupport);
		print OUT "aveCladeSizeSoftSupport\t";
	print OUT "NumHardContra\t";
		print OUT "aveSupportHardConflict\t" if (@STsupport);
		print OUT "aveCladeSizeHardConflict\t";
	print OUT "NumSoftContra\t";
		print OUT "aveSupportSoftConflict\t" if (@STsupport);
		print OUT "aveCladeSizeSoftConflict\t";
	print OUT "NumEquiv\t";
		print OUT "aveSupportEquiv\t" if (@STsupport);
		print OUT "aveCladeSizeEquiv\n";
	
	print NODE "\t" if (@STsupport);
	print NODE "\t\t\t\t\tQSindex(clade)";
		print NODE "\t\tStudent t (p-value)" if ($statAnalysis);
		print NODE "\n";
	print NODE "Replicate\tNode\tCladeID\tCladeSize\t";
	print NODE "CladeSupport\t" if (@STsupport);
	print NODE "Status\tmean\tvariance\t";
	print NODE "QS < +1\tQS not 0\tQS > -1\t" if ($statAnalysis);
	print NODE "NumSourceHardMatch\tNumSourceHardMismatch\tNumSourceEquivocal\t";
		print NODE "NumSourceSoftMatch\tNumSourceSoftMismatch\t" unless ($reducedAnalysis);
	for (my $sourceCount = 1; $sourceCount <= $NumSource; $sourceCount++)
		{
		print NODE "$treeNames{$sourceCount+1}\t";
		}
	print NODE "\n";
	}

print "\nDetermining qualitative clade support ...\n";
	$startTime = time;
	my $superCount = 0;
	my ($NumHardMatch, $NumSoftMatch, $NumHardContra, $NumSoftContra, $NumEquiv) = (0, 0, 0, 0, 0);
	my ($treeQS, $treeQSvar) = (0, 0);
	my ($sumSupportHardMatch, $sumSizeHardMatch) = (0, 0);
	my ($sumSupportSoftMatch, $sumSizeSoftMatch) = (0, 0);
	my ($sumSupportHardContra, $sumSizeHardContra) = (0, 0);
	my ($sumSupportSoftContra, $sumSizeSoftContra) = (0, 0);
	my ($sumSupportEquiv, $sumSizeEquiv) = (0, 0);
	my ($aveSizeHardMatch, $aveSizeSoftMatch, $aveSizeHardContra, $aveSizeSoftContra, $aveSizeEquiv);
	my ($aveSupportHardMatch, $aveSupportSoftMatch, $aveSupportHardContra, $aveSupportSoftContra, $aveSupportEquiv);
	
	foreach my $superNode (@superTree)
		{
		my $status;
		$superCount++;
		printf "\tNode $superCount of %s\n", scalar(@superTree);
		my $sizeST = ($superNode =~ tr/1//);
		my ($NumSourceMatch, $NumSourceMismatch, $NumSourceSoftMatch, $NumSourceSoftMismatch, $NumSourceEquiv) = (0, 0, 0, 0, 0);
		my ($cladeQS, $varCladeQS, $seCladeQS) = (0, 0, 0);
		print "\nST node: $superCount\t$superNode\n\t\t$STcladeID[$superCount-1]\n" if ($verbose);
		for (my $sourceCount = 0; $sourceCount < $NumSource; $sourceCount++)
			{
			my $nodeCount = 0;
			print "\tSource tree: $sourceCount\n" if ($verbose);
			my ($hardMatch, $hardMismatch, $softMatch, $softMismatch) = (0, 0, 0, 0);
			
			#Determine status of supertree clade WRT specific source tree clade (comparison level 1)
				foreach my $sourceNode (@{$sourceTree{$sourceCount}})
					{
					$nodeCount++;
					print "\t\tSource node: $nodeCount\t$sourceNode" if ($verbose);
					if ($superNode eq $sourceNode)	# Exact match between supertree clade on source tree
						{
						$hardMatch = 1;
						print "\t\t\tNode status: Hard match\n" if ($verbose);
						last;	# Search next source tree
						}
					else	#No exact match; compare nodes taxon by taxon
						{
						my ($extraSource, $extraST, $overlap, $extraQuestion, $extraClade) = (0, 0, 0, 0, 0);
						for (my $position = 0; $position < length($superNode); $position++)
							{
							if (substr($superNode, $position, 1) eq "0" and substr($sourceNode, $position, 1) eq "1")
								{
								$extraSource = 1;	#Extra taxon in source tree clade
								}
							if (substr($superNode, $position, 1) eq "1" and substr($sourceNode, $position, 1) eq "0")
								{
								$extraST = 1;	#Extra taxon in supertree clade
								}
							if (substr($superNode, $position, 1) eq "1" and substr($sourceNode, $position, 1) eq "1")
								{
								$overlap++;		#Taxon exists in clades of both trees
								}
							if (substr($superNode, $position, 1) eq "0" and substr($sourceNode, $position, 1) eq "?")
								{
								$extraQuestion = 1;		#Missing taxon outside of supertree clade
								}
							if (substr($superNode, $position, 1) eq "1" and substr($sourceNode, $position, 1) eq "?")
								{
								$extraClade = 1;	#Missing taxon within supertree clade
								}
							}
						if ($overlap)	#Clades on source and supertree have taxa in common
							{
							if ($extraSource and $extraST)	# True regardless if full or reduced analysis
								{
								$hardMismatch = 1;
								print "\t\t\tNode status: Hard mismatch\n" if ($verbose);
								last;	# Search next source tree
								}
							if ($reducedAnalysis)	# Calculate QS values using pruned supertree (ignoring variables involving ?s)
								{
								if (not $extraSource and not $extraST)	# Exact match, ignoring missing taxa
									{
									$hardMatch = 1;
									print "\t\t\tNode status: Hard match\n" if ($verbose);
									last;	# Search next source tree
									}
								}
							else	# Calculate full QS values
								{
								if ($overlap < $sizeST)
									{
									if ($extraST)
										{
										if ($extraQuestion)
											{
											$softMismatch = 1;
											print "\t\t\tNode status: Soft mismatch (missing taxa may contradict supertree clade)\n" if ($verbose);
											}
										else
											{
											print "\t\t\tNode status: N/A (supertree contains source tree clade)\n" if ($verbose);
											}
										}
									if ($extraClade)
										{
										if ($extraSource)
											{
											$softMismatch = 1;
											print "\t\t\tNode status: Soft mismatch (missing taxa may contradict supertree clade)\n" if ($verbose);
											}
										elsif ($extraQuestion)
											{
											print "\t\t\tNode status: Equivocal (missing taxa may both support and contradict supertree clade)\n" if ($verbose);
											}
										else
											{
											$softMatch = 1;
											print "\t\t\tNode status: Soft match (missing taxa will not contradict supertree clade)\n" if ($verbose);
											}
										}
									
									}
								if ($overlap == $sizeST)	#Implies that $extraST = $extraClade = 0
									{
									if ($extraQuestion)
										{
										$softMatch = 1;
										print "\t\t\tNode status: Soft match (missing taxa will not contradict supertree clade)\n" if ($verbose);
										}
									if ($extraSource)
										{
										print "\t\t\tNode status: N/A (source tree contains supertree clade)\n" if ($verbose);
										}
									}
								}
							}
						else	#No taxa in common
							{
							print "\t\t\tNode status: N/A (clades do not overlap)\n" if ($verbose);
							}
						}
					}
	
			# Determine status of supertree clade on source tree (comparison level 2)
				if ($hardMatch)
					{
					$NumSourceMatch++;
					$QSclade{$superCount}{$sourceCount} = "hardMatch";
					print "\n\t\tSource tree status: Hard match ($NumSourceMatch)\n\n" if ($verbose);
					}
				elsif ($hardMismatch)
					{
					$NumSourceMismatch++;
					$QSclade{$superCount}{$sourceCount} = "hardMismatch";
					print "\n\t\tSource tree status: Hard contradiction ($NumSourceMismatch)\n\n" if ($verbose);
					}
				else
					{
					if ($softMatch and not $softMismatch)
						{
						$NumSourceSoftMatch++;
						$QSclade{$superCount}{$sourceCount} = "softMatch";
						print "\n\t\tSource tree status: Soft match; missing taxa will not contradict supertree clade ($NumSourceSoftMatch)\n\n" if ($verbose);
						}
					elsif (not $softMatch and $softMismatch)
						{
						$NumSourceSoftMismatch++;
						$QSclade{$superCount}{$sourceCount} = "softMismatch";
						print "\n\t\tSource tree status: Soft mismatch; missing taxa will not support supertree clade ($NumSourceSoftMismatch)\n\n" if ($verbose);
						}
					else
						{
						$NumSourceEquiv++;
						$QSclade{$superCount}{$sourceCount} = "equivocal";
						print "\n\t\tSource tree status: Equivocal; missing taxa will support and contradict supertree clade ($NumSourceEquiv)\n\n" if ($verbose);
						}
					}
			}
	
		# Determine status of supertree clade among set of source trees (comparison level 3)
			if ($NumSourceMatch)	#Node is directly supported at least once; outweighs all other options
				{
				$status = "hardSupport";
				$NumHardMatch++;
				$sumSupportHardMatch += $STsupport[$superCount-1] if (@STsupport);
		#		$varSupportHardMatch += $STsupport[$superCount-1] * $STsupport[$superCount-1];
				$sumSizeHardMatch += $sizeST;
		#		$varSizeHardMatch += $sizeST * $sizeST;
				}
			else
				{
				if ($NumSourceSoftMatch and not $NumSourceMismatch)		#No direct support, but not contradicted either
					{
					$status = "softSupport";
					$NumSoftMatch++;
					$sumSupportSoftMatch += $STsupport[$superCount-1] if (@STsupport);
			#		$varSupportSoftMatch += $STsupport[$superCount-1] * $STsupport[$superCount];
					$sumSizeSoftMatch += $sizeST;
			#		$varSizeSoftMatch += $sizeST * $sizeST;
					}
				else	#No direct support, but some hard contradictions
					{
					if ($NumSourceMismatch == $NumSource)		#Contradicted by every source tree
						{
						$status = "hardConflict";
						$NumHardContra++;
						$sumSupportHardContra += $STsupport[$superCount-1] if (@STsupport);
				#		$varSupportHardContra += $STsupport[$superCount-1] * $STsupport[$superCount-1];
						$sumSizeHardContra += $sizeST;
				#		$varSizeHardContra += $sizeST * $sizeST;
						}
					elsif ($NumSourceMismatch)		#No direct support and limited number of contradictions
		#			elsif (not $NumHard and $NumSoft and $NumMis)	#No direct support and limited number of contradictions
						{
						$status = "softConflict";
						$NumSoftContra++;
						$sumSupportSoftContra += $STsupport[$superCount-1] if (@STsupport);
				#		$varSupportSoftContra += $STsupport[$superCount-1] * $STsupport[$superCount-1];
						$sumSizeSoftContra += $sizeST;
				#		$varSizeSoftContra += $sizeST * $sizeST;
						}
					else
						{
						$status = "equivocal";
						$NumEquiv++;
						$sumSupportEquiv += $STsupport[$superCount-1] if (@STsupport);
				#		$varSupportEquiv += $STsupport[$superCount-1] * $STsupport[$superCount-1];
						$sumSizeEquiv += $sizeST;
				#		$varSizeEquiv += $sizeST * $sizeST;
						}
					}
				}
	
		$cladeQS = ($NumSourceMatch - $NumSourceMismatch + (0.5 * $NumSourceSoftMatch) - (0.5 * $NumSourceSoftMismatch)) / $NumSource;
		$varCladeQS = $NumSourceMatch + $NumSourceMismatch + (0.25 * $NumSourceSoftMatch) + (0.25 * $NumSourceSoftMismatch);
			$varCladeQS = ($varCladeQS - ($cladeQS**2 / $NumSource)) / ($NumSource - 1);
		$seCladeQS = sqrt($varCladeQS / $NumSource);
	
		$treeQS += $cladeQS;
			$treeQSvar += $cladeQS * $cladeQS;
	
		# Print results for individual clades to $outFile2
			if ($verbose)
				{
				print "\tNumber of source trees: $NumSource\n";
				print "\tStatus among all source trees: $status\n";
				print "\t\tHard match: $NumSourceMatch\n";
				print "\t\tSoft match: $NumSourceSoftMatch\n" unless ($reducedAnalysis);
				print "\t\tHard mismatch: $NumSourceMismatch\n";
				print "\t\tSoft mismatch: $NumSourceSoftMismatch\n" unless ($reducedAnalysis);
				print "\t\tEquivocal: $NumSourceEquiv\n";
				print "\tRunning tally of supertree nodes:\n";
				print "\t\tHard support: $NumHardMatch\n";
				print "\t\tSoft support: $NumSoftMatch\n" unless ($reducedAnalysis);
				print "\t\tHard conflict: $NumHardContra\n";
				print "\t\tSoft conflict: $NumSoftContra\n" unless ($reducedAnalysis);
				print "\t\tEquivocal: $NumEquiv\n";
				}
	
			print NODE "$iRep\t$superCount\t$STcladeID[$superCount-1]\t$sizeST\t";
			print NODE "$STsupport[$superCount-1]\t" if (@STsupport);
			if ($reducedAnalysis)
				{
				print NODE "n/a\t";
				}
			else
				{
				print NODE "$status\t";
				}
			printf NODE "%1.3f\t%1.3f\t", $cladeQS, $varCladeQS;
	
			# Print out t-test results (unless switched off)
				if ($statAnalysis)
					{
					my $df = $NumSource - 1;

					my ($t, $lowerP);
					
					$seCladeQS = 0.00001 if ($seCladeQS == 0);	# Safety in case SE = 0;
					
					# Test that QS < 1 (one-tailed)
						$t = ($cladeQS - 1) / $seCladeQS;
						$lowerP = _subtprob($df, $t);
						printf NODE "%.3f (%.4f)\t", $t, $lowerP;
		
					# Test that QS is not equal to 0 (two-tailed)
						$t = ($cladeQS - 0) / $seCladeQS;
						$lowerP = _subtprob($df, $t);
							$lowerP = 1 - $lowerP if ($t < 0);
							$lowerP *= 2;	# For two-tailed testing
						printf NODE "%.3f (%.4f)\t", $t, $lowerP;
		
					# Test that QS > -1 (one-tailed)
						$t = ($cladeQS - -1) / $seCladeQS;
						$lowerP = _subtprob($df, $t);
						printf NODE "%.3f (%.4f)\t", $t, $lowerP;
					}
	
			print NODE "$NumSourceMatch\t$NumSourceMismatch\t$NumSourceEquiv\t";
				print NODE "$NumSourceSoftMatch\t$NumSourceSoftMismatch\t" unless ($reducedAnalysis);
			for (my $sourceCount = 0; $sourceCount < $NumSource; $sourceCount++)
				{
				print NODE "$QSclade{$superCount}{$sourceCount}\t";
				}
			print NODE "\n";
			
			if ($reducedAnalysis)
				{
				$nodeLabel{$superCount - 1} = sprintf "'%1.3f'", $cladeQS;
				}
			else
				{
				$nodeLabel{$superCount - 1} = sprintf "'$abbrevStatus{$status}_%1.3f'", $cladeQS;
				}		
		}
	$totalTime = time - $startTime;
	print "\tTime taken: $totalTime seconds\n";

# Determine results over tree as a whole
	if ($NumHardMatch > 1)
		{
		$aveSupportHardMatch = $sumSupportHardMatch / $NumHardMatch if (@STsupport);
		$aveSizeHardMatch = $sumSizeHardMatch / $NumHardMatch;
		}
	if ($NumHardMatch == 1)
		{
		$aveSupportHardMatch = $sumSupportHardMatch if (@STsupport);
		$aveSizeHardMatch = $sumSizeHardMatch;
		}
	if (not $NumHardMatch)
		{
		$aveSupportHardMatch = "n/a";
		$aveSizeHardMatch = "n/a";
		}
	
	if ($NumSoftMatch > 1)
		{
		$aveSupportSoftMatch = $sumSupportSoftMatch / $NumSoftMatch if (@STsupport);
		$aveSizeSoftMatch = $sumSizeSoftMatch / $NumSoftMatch;
		}
	if ($NumSoftMatch == 1)
		{
		$aveSupportSoftMatch = $sumSupportSoftMatch if (@STsupport);
		$aveSizeSoftMatch = $sumSizeSoftMatch;
		}
	if (not $NumSoftMatch)
		{
		$aveSupportSoftMatch = "n/a";
		$aveSizeSoftMatch = "n/a";
		}
	
	if ($NumHardContra > 1)
		{
		$aveSupportHardContra = $sumSupportHardContra / $NumHardContra if (@STsupport);
		$aveSizeHardContra = $sumSizeHardContra / $NumHardContra;
		}
	if ($NumHardContra == 1)
		{
		$aveSupportHardContra = $sumSupportHardContra if (@STsupport);
		$aveSizeHardContra = $sumSizeHardContra;
		}
	if (not $NumHardContra)
		{
		$aveSupportHardContra = "n/a";
		$aveSizeHardContra = "n/a";
		}
	
	if ($NumSoftContra > 1)
		{
		$aveSupportSoftContra = $sumSupportSoftContra / $NumSoftContra if (@STsupport);
		$aveSizeSoftContra = $sumSizeSoftContra / $NumSoftContra;
		}
	if ($NumSoftContra == 1)
		{
		$aveSupportSoftContra = $sumSupportSoftContra if (@STsupport);
		$aveSizeSoftContra = $sumSizeSoftContra;
		}
	if (not $NumSoftContra)
		{
		$aveSupportSoftContra = "n/a";
		$aveSizeSoftContra = "n/a";
		}
	
	if ($NumEquiv > 1)
		{
		$aveSupportEquiv = $sumSupportEquiv / $NumEquiv if (@STsupport);
		$aveSizeEquiv = $sumSizeEquiv / $NumEquiv;
		}
	if ($NumEquiv == 1)
		{
		$aveSupportEquiv = $sumSupportEquiv if (@STsupport);
		$aveSizeEquiv = $sumSizeEquiv;
		}
	if (not $NumEquiv)
		{
		$aveSupportEquiv = "n/a";
		$aveSizeEquiv = "n/a";
		}
	
	$treeQS /= $superCount;
	$treeQSvar = ($treeQSvar - ($treeQS**2 / scalar(@superTree))) / (scalar(@superTree) - 1);
	
	if ($iRep > 1)
		{
		print OUT "$iRep\t";
		}
	else
		{
		print OUT "overall\t";
		}
	
	print OUT "$superCount\t";
	printf OUT "%1.3f\t%1.3f\t", $treeQS, $treeQSvar;
	print OUT "$NumHardMatch\t";
		if ($NumHardMatch)
			{
			printf OUT "%3.3f\t", $aveSupportHardMatch if (@STsupport);
			printf OUT "%3.3f\t", $aveSizeHardMatch;
			}
		else
			{
			print OUT "$aveSupportHardMatch\t" if (@STsupport);
			print OUT "$aveSizeHardMatch\t";
			}
	print OUT "$NumSoftMatch\t";
		if ($NumSoftMatch)
			{
			printf OUT "%3.3f\t", $aveSupportSoftMatch if (@STsupport);
			printf OUT "%3.3f\t", $aveSizeSoftMatch;
			}
		else
			{
			print OUT "$aveSupportSoftMatch\t" if (@STsupport);
			print OUT "$aveSizeSoftMatch\t";
			}
	print OUT "$NumHardContra\t";
		if ($NumHardContra)
			{
			printf OUT "%3.3f\t", $aveSupportHardContra if (@STsupport);
			printf OUT "%3.3f\t", $aveSizeHardContra;
			}
		else
			{
			print OUT "$aveSupportHardContra\t" if (@STsupport);
			print OUT "$aveSizeHardContra\t";
			}
	print OUT "$NumSoftContra\t";
		if ($NumSoftContra)
			{
			printf OUT "%3.3f\t", $aveSupportSoftContra if (@STsupport);
			printf OUT "%3.3f\t", $aveSizeSoftContra;
			}
		else
			{
			print OUT "$aveSupportSoftContra\t" if (@STsupport);
			print OUT "$aveSizeSoftContra\t";
			}
	print OUT "$NumEquiv\t";
		if ($NumEquiv)
			{
			printf OUT "%3.3f\t", $aveSupportEquiv if (@STsupport);
			printf OUT "%3.3f\n", $aveSizeEquiv;
			}
		else
			{
			print OUT "$aveSupportEquiv\t" if (@STsupport);
			print OUT "$aveSizeEquiv\n";
			}
	
close OUT;
close NODE;

# Perform bootstrap resampling of source trees to determine confidence intervals on QS index
	if ($bootRep > 1)
		{
		print "\nRunning $bootRep bootstrap replicates ... \n";
		my $timeZero = time;
		my (%aveBootQS, %varBootQS, %seBootQS);
		my ($aveTreeBootQS, $varTreeBootQS, $seTreeBootClade);

		open (BOOT, $write, $outFile4) or die "Cannot open output file $outFile4\n";
			print BOOT "\t" if (@STsupport);
			print BOOT "\t\tQSindex(clade)\t\tStudent t (p-value)\n";
			print BOOT "Node\tCladeID\t";
			print BOOT "CladeSupport\t" if (@STsupport);
			print BOOT "mean\tvariance\t";
			print BOOT "QS < +1\tQS not 0\tQS > -1\n";

		# Do bootstrap resampling
			# Done in two loops: 1) stores counts for each source tree, 2) generates QS values using these counts
			# Will be faster in most cases than direct incrementing of QS values in a single loop (running time: $NumSource * $superCount)
				# Worst-case running time = $NumSource + ($NumSource * $superCount) when each source tree sampled once
				#                         = $NumSource * (1 + $superCount)
				# Best-case running time = $NumSource + (1 * $superCount) when one source tree sampled
				#                        = $NumSource + $superCount

			for (1..$bootRep)
				{
				print "\tReplicate number: $_\n" if (int($_/100)*100 == $_ and $verbose);
				my %sourceTreeCount;
				my $treeRepQS;
				for (1..$NumSource)
					{
					my $randTree = int(rand($NumSource));
					$sourceTreeCount{$randTree}++;
					}
				for (my $superClade = 1; $superClade <= $superCount; $superClade++)
					{
					my ($bootSourceMatch, $bootSourceMismatch, $bootSourceSoftMatch, $bootSourceSoftMismatch, $bootSourceEquiv) = (0, 0, 0, 0, 0);
					for (my $sourceCount = 0; $sourceCount < $NumSource; $sourceCount++)
						{
						next unless (defined $sourceTreeCount{$sourceCount});
						# Read status of supertree clade on source tree (comparison level 2)
							$bootSourceMatch += $sourceTreeCount{$sourceCount} if ($QSclade{$superClade}{$sourceCount} eq "hardMatch");
							$bootSourceMismatch += $sourceTreeCount{$sourceCount} if ($QSclade{$superClade}{$sourceCount} eq "hardMismatch");
							$bootSourceSoftMatch += $sourceTreeCount{$sourceCount} if ($QSclade{$superClade}{$sourceCount} eq "softMatch");
							$bootSourceSoftMismatch += $sourceTreeCount{$sourceCount} if ($QSclade{$superClade}{$sourceCount} eq "softMismatch");
							$bootSourceEquiv += $sourceTreeCount{$sourceCount} if ($QSclade{$superClade}{$sourceCount} eq "equivocal");
						}
					# Determine status of supertree clade among set of source trees (comparison level 3)
						my $repQS = ($bootSourceMatch - $bootSourceMismatch + (0.5 * $bootSourceSoftMatch) - (0.5 * $bootSourceSoftMismatch)) / $NumSource;
#print "$repQS\n";
						$aveBootQS{$superClade} += $repQS;
						$varBootQS{$superClade} += $repQS * $repQS;
						$treeRepQS += $repQS;
					}
				$treeRepQS /= $superCount;
#print "\t$treeRepQS\n";
				$aveTreeBootQS += $treeRepQS;
				$varTreeBootQS += $treeRepQS * $treeRepQS;
				}
			$aveTreeBootQS /= $bootRep;
			$varTreeBootQS = ($varTreeBootQS - ($aveTreeBootQS**2 / $bootRep)) / ($bootRep - 1);
		
		# Get average QS values over all bootstrap replicates
			for (my $superClade = 1; $superClade <= $superCount; $superClade++)
				{
				$aveBootQS{$superClade} /= $bootRep;
				$varBootQS{$superClade} = ($varBootQS{$superClade} - ($aveBootQS{$superClade}**2 / $bootRep)) / ($bootRep - 1);
					$seBootQS{$superClade} = sqrt($varBootQS{$superClade} / $bootRep);

				print BOOT "$superClade\t$STcladeID[$superClade-1]\t";
				print BOOT "$STsupport[$superClade-1]\t" if (@STsupport);
				printf BOOT "%1.3f\t%1.3f\t", $aveBootQS{$superClade}, $varBootQS{$superClade};

				# Print out t-test results
					my $df = $NumSource - 1;
					
					my ($t, $lowerP);
					
					$seBootQS{$superClade} = 0.00001 if ($seBootQS{$superClade} == 0);	# Safety in case SE = 0;
					
					# Test that QS < 1 (one-tailed)
						$t = ($aveBootQS{$superClade} - 1) / $seBootQS{$superClade};
						$lowerP = _subtprob($df, $t);
						printf BOOT "%.3f (%.4f)\t", $t, $lowerP;
		
					# Test that QS is not equal to 0 (two-tailed)
						$t = ($aveBootQS{$superClade} - 0) / $seBootQS{$superClade};
						$lowerP = _subtprob($df, $t);
							$lowerP = 1 - $lowerP if ($t < 0);
							$lowerP *= 2;	# For two-tailed testing
						printf BOOT "%.3f (%.4f)\t", $t, $lowerP;
		
					# Test that QS > -1 (one-tailed)
						$t = ($aveBootQS{$superClade} - -1) / $seBootQS{$superClade};
						$lowerP = _subtprob($df, $t);
						printf BOOT "%.3f (%.4f)\n", $t, $lowerP;
				}
		
		print BOOT "QS tree\t\t";
		print BOOT "\t" if (@STsupport);
		printf BOOT "%1.3f\t%1.3f\t", $aveTreeBootQS, $varTreeBootQS;
		my $bootTime = time - $timeZero;
		print "\tTime taken: $bootTime seconds\n";
		}

# Print nexus-formatted tree file of supertree with QS index labels
	open (TREE, ">$outFile3") or die "Cannot open output file $outFile3 to write tree description.\n";
		print TREE "#NEXUS\n\nbegin trees;\n";
		my ($position, $CurrentClade) = (0, 0);
		undef %open;
		undef %ParentClade;
		my $CladeID = -1;
		my $newST;
		my $OldCurrent;
		while ($position < length($AllTrees[0]))
			{
			my $offset = 1;
			if (substr($AllTrees[0], $position, 1) eq "(")		# New clade; go one level deeper
				{
				$CladeID++;
				$ParentClade{$CladeID} = $CurrentClade;
				$CurrentClade = $CladeID;
				$newST .= "(";
				}
			elsif (substr($AllTrees[0], $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$open{$CurrentClade} = 0;
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
#				$stop[$OldCurrent] = $position;
				$newST .= ")";
				if ($nodeLabel{$OldCurrent})	# Do not label root node with MRP outgroup
					{
					$newST .= "$nodeLabel{$OldCurrent}";
					}
				}
			else
				{
				$newST .= substr($AllTrees[0], $position, 1);
				}
			$position++;
			}
		print TREE "\tTREE QS_supertree = $newST;\n";
		print TREE "end;\n";
	close TREE;

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "DOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "Mac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "Unix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}
	
sub _subtprob	# Calculates upper probability from Student's t-distribution; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my ($n, $x) = @_;

	my ($a,$b);
	my $w = atan2($x / sqrt($n), 1);
	my $z = cos($w) ** 2;
	my $y = 1;

	for (my $i = $n-2; $i >= 2; $i -= 2)
		{
		$y = 1 + ($i-1) / $i * $z * $y;
		} 

	if ($n % 2 == 0)
		{
		$a = sin($w)/2;
		$b = .5;
		}
	else
		{
		$a = ($n == 1) ? 0 : sin($w)*cos($w)/pi;
		$b= .5 + $w/pi;
		}
	return max(0, 1 - $b - $a * $y);
	}
	
sub max	# Finds maximum value in an array; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my $max = shift;
	my $next;
	while (@_)
		{
		$next = shift;
		$max = $next if ($next > $max);
		}	
	return $max;
	}

# Version history
#
#	v1.2.1 (April 18, 2007)
#		- uses (standard) Math::Trig module to get values for pi
#		- bug fix relating to how variances were calculated
#		- checks that necessary input files exist before proceeding
#		- fixed bug in tree printing routine such that the wrong values were being associated
#		  with the wrong nodes (thanks to Rob Davis for spotting this)
#		- minor bug fixes
#
#	v1.2 (August 3, 2005)
#		- major bug fix to calculation of p-values for Student's t-distribution (no longer uses
#		  gamma function and Lanczos approximation of it)
#		- added possibility to do bootstrap resampling of source trees to determine confidence
#		  limits on (r)QS values; end result: (r)QS values are roughly identical, but confidence
#		  intervals are much smaller
#		- added ability to perform rQS analysis where supertree and source tree are functionally
#		  pruned to same taxon set; now also represents default setting
#		- now outputs QS status of each source tree for each node to file *_(r)QS_clade.txt
#		- minor change to output files; now named using base name from input file
#		- bug fix relating to how variances were calculated
#		- added explicit GNU GPL statement
#		- checks that necessary input files exist before proceeding
#		- minor bug fixes and code cleaning
#
#	v1.1 (April 6, 2004)
#		- added ability to test whether the QS value for a given node is significantly different from
#		  any of +1, 0, or -1 using a Student's t-test
#		- QS_clade.txt no longer an optional output file
#		- supertree now also output as a nexus-formatted treefile with QS values listed as node labels
#		- added ability to detect and accomodate different styles of line breaks
#		- minor bug fixes and code cleaning
#
#	v1.0.4 (January 9, 2004)
#		- minor bug fixes and code cleaning
#
#	v1.0.3 (August 19, 2002)
#		- minor bug fixes and code cleaning
#
#	v1.0.2 (August 1, 2002)
#		- minor bug fixes and code cleaning
#
#	v1.0.1 (July 30, 2002)
#		- minor bug fixes and code cleaning
#
#	v1.0 (July 29, 2002)
#		- initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
